#-----------------------------------------------------#
# example.R
#
# Illustrates use of jackknife code 'jregress.R'.
# Implements methods from "Jackknife standard errors for clustered regression"
# by Bruce E. Hansen
#
# This example file uses the pre-processed data file Meng_1.dta, and uses the R package haven to load the data.
#
# Six example regressions are shown. All report jackknife standard errors.
#
# The first example calculates the regression for the model corresponding to the first column of Table 7 of Meng, Qian, and Yared (2015), but with jackknife standard errors clustered by province, and inference (p-values and condfidence intervals) calculated by the adjusted t distribution.
#
# The second example repeats this regression but omits clustering.
#
# The third example repeats the first regression and displays the adjustment coefficients.
#
# The fourth example repeats the first regression but calculates inference by the standard t distribution.
#
# The fifth example takes the first regression and adds cluster-level fixed effects.
#
# The sixth example adds two cluster-level dummy variable indicators to the first regression. The dummy variables fail clusterwise invertibility. The regression and standard errors are reported, but a warning message is displayed.
# 
# 
#-----------------------------------------------------#

library(haven)
source('jregress.R')

dat1 <- read_dta("Meng_1.dta")
dat1 <- subset(dat1,dat1$lgrain_pred_famdum5860!='.')
dat1 <- subset(dat1,dat1$year>=1953)
dat1 <- subset(dat1,dat1$year<=1982)
dat1$i.year <- as.factor(dat1$year)
dat1$diff_pred <- dat1$lgrain_pred_famdum5860 - dat1$lgrain_pred
dat1$Beijing <- (dat1$province == "Beijing")
dat1$Tianjin <- (dat1$province == "Tianjin")

cat("\n-------------------------------------------------------------\n")
cat("Example 1\n")
jack <- jregress(ldeath_b~diff_pred+lgrain_pred+ltotpop+lurbpop+i.year, data=dat1, clusters=prov)

cat("\n-------------------------------------------------------------\n")
cat("Example 2\n")
jack <- jregress(ldeath_b~diff_pred+lgrain_pred+ltotpop+lurbpop+i.year, data=dat1)

cat("\n-------------------------------------------------------------\n")
cat("Example 3\n")
jack <- jregress(ldeath_b~diff_pred+lgrain_pred+ltotpop+lurbpop+i.year, data=dat1, clusters=prov, df.display=TRUE)

cat("\n-------------------------------------------------------------\n")
cat("Example 4\n")
jack <- jregress(ldeath_b~diff_pred+lgrain_pred+ltotpop+lurbpop+i.year, data=dat1, clusters=prov, df.adjust=FALSE)

cat("\n-------------------------------------------------------------\n")
cat("Example 5\n")
jack <- jregress(ldeath_b~diff_pred+lgrain_pred+ltotpop+lurbpop+i.year, data=dat1, clusters=prov, fe=TRUE)

cat("\n-------------------------------------------------------------\n")
cat("Example 6\n")
jack <- jregress(ldeath_b~diff_pred+lgrain_pred+ltotpop+lurbpop+Beijing+Tianjin+i.year, data=dat1, clusters=prov)
