/* -------------------------------------------------------------------------- */
// Stata for Hansen (2024): Jackknife Standard Errors for Clustered Regression 
/* -------------------------------------------------------------------------- */
*! version 0.10.1 02aug2024


/* 		
jregress depvar [ indvars ] [ if ] [ in ] [ , options]

Examples:

(1) Simple linear regression of y on x1
jregress y x1

(2) Regression of y on x1, x2, indicators for categorical variable a, and standard errors clustered by id
jregress y x1 x2 i.a, cluster(id)

(3) Add the interaction between continuous variable x2 and a
jregress y x1 c.x2##i.a, cluster(id) 

(4) Estimate the model for observations where v1 is greater than zero
jregress y x1 x2 i.a if v1>0, cluster(id)

(5) Add cluster-level fixed effects to the model
jregress y x1 x2 i.a, cluster(id) fe

(6) Display degree-of-freedom adjustment coefficients
jregress y x1 x2 i.a, cluster(id) dfdisplay 

(7) Do not adjust student t degree-of-freedom
jregress y x1 x2 i.a, cluster(id) noadjust

*/


program define jregress, eclass sortpreserve
	version 18.0
	
	syntax varlist(numeric ts fv) [if] [in] [, CLuster(string) vce(string) 		 		/// 			// vce options
												noCONStant FE noADJust 					///				// additional options
												noHEader noTABle noBLock DFDISPlay		///				// display options
												*										///
											]
	marksample touse
	
	
	// --- Parse syntax --- //
	
	_get_diopts diopts, `options'	// ensure only display options included in '*'
	
	gettoken depvar indepvars : varlist
	_fv_check_depvar `depvar'
	_rmcoll `indepvars' if `touse', `constant' expand // flag (but not remove) collinear regressors; Q: What if collinear w/ fixed effects? \\\
																				// A: Default is omit collinear variable w/o message, as in xtreg.
	local cnames `r(varlist)'
	
	
	// construct vector containing flagged/omitted variables
	local k : word count `cnames'
	if "`constant'" == "" & "`fe'" == "" {
		local k = `k' + 1
		local cons _cons
	}
	tempname m_omit omit
	matrix `m_omit' = J(1,`k',0)
	matrix colnames `m_omit' = `cnames' `cons'
	_ms_omit_info `m_omit'
	matrix `omit' = r(omit)
	
	
	_vce_parse `touse', argoptlist(CLuster) old : , vce(`vce') cluster(`cluster') // parse vce w/ backward compatibility
	local vce 			"`r(vce)'"
	local clustervar 	"`r(cluster)'"
	
	
	
	// --- Error checking --- //
	
	if "`clustervar'" != "" { // non-numeric cluster variable
		capture confirm numeric variable `clustervar'
		if _rc {
			display as error "cluster variable {bf:`clustervar'} is non-numeric"
			exit(198)
		}
		
		sort `clustervar' //sort by clustervar for calculation
	}
	
	
	if "`fe'" == "fe" { // errors specific to cluster fixed effects option
		* specified fe option w/o defining cluster variable
		if "`clustervar'" == "" {
			display as error "{bf: `fe'} option specified without a cluster variable"
			exit(198)
		}
		
		* included cluster variable as (in)dependent variable (note: `clustervar' != "" redundant given previous check)
		if `: list clustervar in varlist' { // used varlist instead of indepvars to include dependent variable case. Q: Separate checks?
			display as error "cluster variable {bf:`clustervar'} cannot be included as an (in)dependent variable when {bf:`fe'} option specified"
			exit(198)
		}
	}
	
	
	
	// --- Estimation --- //
	
	tempname b V N G G_vec countg rank df_r K scale CV R2
	
	mata: jreg("`depvar'", "`cnames'", "`touse'", "`omit'", ///
		"`vce'", "`clustervar'", "`constant'", "`fe'", "`adjust'", 	///
		"`b'", "`V'", "`N'", "`G'", "`G_vec'", "`countg'", "`rank'", "`df_r'", ///
		"`K'", "`scale'", "`CV'", "`R2'")
	
	if "`constant'" == "" & "`fe'" == "" {
		local cnames "`cnames' _cons"
	}
	
	matrix colnames `b' 		= `cnames'
	matrix colnames `V' 		= `cnames'
	matrix rownames `V' 		= `cnames'
	
	
	
	// --- Store Results --- //
	
	ereturn post `b' `V', esample(`touse') buildfvinfo
	
	if "`adjust'" == "" { // B-M specific output
		matrix colnames `K' 	= `cnames'
		matrix colnames `scale'		= `cnames'
		ereturn matrix K 		= `K'
		ereturn matrix scale 	= `scale'
	}
	
	if "`vce'" == "cluster" {
		ereturn scalar N_clust = `G'
	}
	ereturn scalar N 		= `N'
	ereturn scalar df_r		= `df_r'
	ereturn scalar rank 	= `rank'
	ereturn scalar r2		= `R2'
	ereturn scalar CV		= `CV'
	
	ereturn local cmdline		`"jregress `0'"'
	ereturn local vce			"`vce'"
	ereturn local vcetype		"Jackknife"
	ereturn local clustervar 	"`clustervar'"
	ereturn local depvar		"`depvar'"
	ereturn local cmd			"jregress"

	
	
	// --- Display output --- // 
	local diopts1 `header' `table' `block' `dfdisplay' `fe' `adjust'
	
	jreg_Display, clustinfo(`G_vec') clustnid(`countg') `diopts' `diopts1'

end





mata:

void jreg(string scalar depvar,		string scalar indepvars,	string scalar touse,
			string scalar omit,		string scalar vcetype,		string scalar clustervar,
			string scalar constant,	string scalar fe,			string scalar adjust,
			string scalar bname,	string scalar Vname,		string scalar nname,
			string scalar gname,	string scalar gvname,		string scalar cgname,
			string scalar rname,	string scalar dfrname,		string scalar Kname,
			string scalar aname,	string scalar cvname,		string scalar r2name)
{
	// Declare objects...
	real vector Y, XY, beta, e, id, m_id, Ymean, Xmean, Yg, betag, eg, ng, Gv, uniq_c, g_id, omit_v
	real matrix X, XX, Q, c_id, Xg, XXg, Qg, Vhat, Betag
	real scalar n, k, G, dfr, rank, rankg, countg, CV, idc, SSE, SST, R2, c
	
	
	// Adjustment objects...
	real vector a, K, Sr, S2r, trL, trLL
	real matrix S, S2, U, V, Ug, Vg, Ur, Vr, Wr, UU, UV, VV, UUXX
	
	
	// Variable assignment/manipulation
	Y = st_data(., depvar, touse)
	X = st_data(., indepvars, touse)
	n = rows(X)
	
	if (constant == "" & fe == ""){
		X = X,J(n,1,1)
	}

	k = cols(X)
	G = n
	
	if (vcetype == "cluster"){ // pre-proccessing for cluster calculations
		id = st_data(., clustervar, touse) // cluster variable
		c_id = panelsetup(id,1) // panel setup for variance calculations

		G = rows(c_id) // no. of clusters
		ng = (c_id[.,2] - c_id[.,1]) :+ 1 // no. obs. per cluster
		Gv = (min(ng),mean(ng),max(ng)) // store cluster info for reporting
		
		// Within Transformation for cluster fixed effects
		if (fe == "fe"){
			m_id = J(n,1,.)
			m_id[c_id[.,1],] = J(G,1,1)
			m_id = quadrunningsum(m_id)	// ordinal cluster id
			Ymean = panelsum(Y,c_id) :/ ng
			Xmean = panelsum(X,c_id) :/ ng
			Y = Y - Ymean[m_id,1]
			X = X - Xmean[m_id,.]
		}
	}
	
	
	// OLS coefficients and residuals
	XX = quadcross(X,X)
	XY = quadcross(X,Y)
	Q = invsym(XX)
	beta = Q*XY
	e = Y - X*beta
	
	// Rank & residual dof
	rank = k - diag0cnt(Q) // each collinear variable corresponds to zero on diagonal of generalized symmetric inverse
	dfr = n - rank
	if (vcetype == "cluster"){
		dfr = G-1 
	}
	
	// R-squared
	SSE = quadsum(e:^2)
	if (constant == "" & fe == ""){
		SST = quadsum((Y :- mean(Y)):^2)
		c = (n-1)/dfr
	}
	else {
		SST = quadsum(Y:^2)
		c = n/dfr
	}
	R2 = 1 - (SSE/SST)
	
	
	// --- VCE estimation --- //
	
	CV = 0
	countg = 0
	U = V = J(G,k^2,0)
	S = Betag = J(G,k,0)
	
	for (g=1; g<=G; g++){ 
		if (vcetype == "cluster"){
			Xg = panelsubmatrix(X,g,c_id)
			Yg = panelsubmatrix(Y,g,c_id)
		}
		else {
			Xg = X[g,.]
			Yg = Y[g,1]
		}
		XXg = quadcross(Xg,Xg)
		Qg = pinv(XX - XXg, rankg)
		countg = countg + (rankg < rank)												
		betag = Qg*(XY - quadcross(Xg,Yg))	// leave-one-out coefficients
		Betag[g,.] = (betag - beta)' 
		eg = Yg - Xg*betag
		CV = CV + quadsum(eg:^2) 			
		Ug = Qg*XXg*Q
		Vg = XXg*(Q + Ug)
		U[g,.] = vec(Ug)'
		V[g,.] = vec(Vg)'
		S[g,.] = quadcolsum((Q + Ug):*Vg)
	}
	
	omit_v = 1 :- st_matrix(omit)		// ... ad hoc correction for omitted variables (3/12/24)
	Betag = Betag :* omit_v
	Vhat = quadcross(Betag,Betag)		// Jackknife variance estimator
	

	
	// --- Degrees of Freedom Adjustment --- //
	
	if (adjust == ""){ 
		
		trL = trLL = J(1,k,0)
		for (r=1; r<=k; r++){		// by coefficient
			
			Sr = S[.,r]
			Ur = U[.,((r-1)*k+1)..(r*k)]
			Vr = V[.,((r-1)*k+1)..(r*k)]
			Wr = Ur :* Sr
			UU = quadcross(Ur,Ur)
			UV = quadcross(Ur,Vr)
			VV = quadcross(Vr,Vr)
			UUXX = UU*XX

			trL[r] = sum(Sr) - trace(UV) 
			trLL[r] = sum(Sr:^2) + trace(UUXX,UUXX) + 2*trace(UV,UV) ///
				  - 2*trace(Vr,Wr,1) - 4*trace(UUXX,UV) + 2*trace(UU,VV) 

		}
		a = sqrt(trL:/diagonal(Q)') 
		K = (trL:^2):/trLL
	}
	
	
	
	// return output
	st_matrix(bname, beta')
	st_matrix(Vname, Vhat)
	st_numscalar(nname,  n)
	st_numscalar(gname,  G)
	st_matrix(gvname, Gv')
	st_numscalar(cgname, countg)
	st_numscalar(rname, rank)
	st_numscalar(dfrname, dfr)
	st_matrix(Kname, K)
	st_matrix(aname, a)
	st_numscalar(cvname, CV)
	st_numscalar(r2name, R2)
}

end





program jreg_Display
	syntax [,									///
			noHEader noTABle noBLock DFDISPlay 	///	// display options
			FE noADJust							/// // estimation options
			clustinfo(string) clustnid(string)	/// // cluster information
			*									///
	]
	_get_diopts diopts, `options'	// ensure only display options included in '*'
	
	
	// Output header
	if "`header'" == "" { // Display header ---> See if we can use SMCL to eliminate use of hard-coded column numbers.
	
		if "`fe'" == "fe" { // different title if fe
			display _newline as text "Fixed-effects (within) regression" _col(42) "Number of obs" _col(65) " = " as result _col(70) %9.0g `e(N)'
		}
		else {
			display _newline as text "Linear regression" _col(42) "Number of obs" _col(65) " = " as result _col(70) %9.0g `e(N)'
		}
	
		if "`e(vce)'" == "cluster" { // cluster display options
			display as text "Cluster variable: `e(clustervar)'" _col(42) "Number of clusters" _col(65) " = " as result _col(70) %9.0g `e(N_clust)'
			display as text _col(42) "R-squared" _col(65) " = " 	as result _col(70) %9.0g `e(r2)'									// Q: Add "within" label when fe specified, as in xtreg?
		
			if "`block'" == "" { // display block containing info on no. of obs. per cluster
				display as text "Obs per cluster:" 	as text _col(42) "Cluster CV criterion" _col(65) " = " as result _col(70) %9.0gc `e(CV)'
				display as text _col(5) "min = " as result %9.0g `clustinfo'[1,1] 	as text _col(42) 
				display as text _col(5) "avg = " as result %9.0g `clustinfo'[2,1] 	
				display as text _col(5) "max = " as result %9.0g `clustinfo'[3,1]  _newline
			}
			else {
				display as text _col(42) "Cluster CV criterion" _col(65) " = " 		as result _col(70) %9.0gc 	`e(CV)' _newline
			}
		
		}
		else {
			display as text _col(42) "R-squared" _col(65) " = " 		as result _col(70) %9.0g 	`e(r2)'
			display as text _col(42) "Cluster CV criterion" _col(65) " = " 		as result _col(70) %9.0gc 	`e(CV)' _newline
		}
	}
	

	// Output table
	
	if "`table'" == "" { // Display coef. table
	
		// Warning message for non-identifed jackknife coefficients
		if `clustnid' > 0 {
			if "`e(vce)'" == "cluster" {
				display as error "{bf:Warning}: Jackknife coefficients not identified in " as result `clustnid' as error "/"`e(N_clust)' " clusters."
			}
			else {
				display as error "{bf:Warning}: Jackknife coefficients not identified in " as result `clustnid' as error "/"`e(N)' " iterations."
			}
			display as error "Jackknife standard errors may be conservative. Recommend respecification of regression model." _newline
		}
		
		// Display table
		if "`adjust'" == ""{ //... adjusted critical value
			* Custom coefficient table subroutine
			_bm_coeftab, dfmatrix(e(K)) scmatrix(e(scale)) `dfdisplay' `diopts'
	
		} 
		else {
			if "`dfdisplay'" == "dfdisplay" { // warning if both `noadjust' and `dfdisplay' specified.
				display as error "{bf:Warning}: Both {bf:`adjust'} and {bf:`dfdisplay'} options specified; {bf:`dfdisplay'} option ignored." _newline
			}
			* Built-in coefficient table function
			_coef_table, `diopts'
		}
		
		// Standard error message
		if "`e(vce)'" == "cluster" {
			display as text "Standard errors calculated by the delete-cluster jackknife."
		} 
		else{
			display as text "Standard errors calculated by the jackknife."
		} 
		
		// (Exact) inference message
		if "`adjust'" == "" {
			display as text "p-values and confidence intervals calculated by the adjusted student t distribution."
		}
		else {
			if "`e(vce)'" == "cluster" {
				display as text "p-values and confidence intervals calculated by the student t(G-1) distribution."
			}
			else {
				display as text "p-values and confidence intervals calculated by the student t(n-k) distribution."
			}
		}
	}
end





program _bm_coeftab, rclass
	syntax, DFMATrix(string) SCMATrix(string) [, level(int `c(level)') DFDISPlay *]
		
	_get_diopts diopts, `options'	// ensure only display options included in '*'
		
	tempname mytab outmat scale df crit t p ll ul
		
	if "`dfdisplay'" == "dfdisplay" { // table formatting
		.`mytab' = ._tab.new, col(9) lmargin(0) ignore(.b)
		.`mytab'.width    13   | .     .     8	   8     8     8     .     .
		.`mytab'.titlefmt  .     .     .   %6s     .     .     .  %24s     .
		.`mytab'.pad       .     2     1     .     .     .     2     3     3
		.`mytab'.numfmt    . %9.0g %9.0g %7.2f %7.3g %7.3f %5.3f %9.0g %9.0g
	}
	else {
		.`mytab' = ._tab.new, col(7) lmargin(0) ignore(.b)
		.`mytab'.width    13   |12    12     8     8    12    12
		.`mytab'.titlefmt  .     .     .   %6s     .  %24s     .
		.`mytab'.pad       .     2     1     .     2     3     3
		.`mytab'.numfmt    . %9.0g %9.0g %7.2f %5.3f %9.0g %9.0g
	}
		
        
		

	local namelist : colname e(b)
    local k : word count `namelist'
	local cns `: list _cons in namelist' // indicator if model includes constant
		
		
	matrix `outmat' = J(9,`k',.)
	matrix colnames `outmat' = `namelist'
	matrix rownames `outmat' = "b" "se" "t" "pvalue" "ll" "ul" "df" "crit" "scale"
		
	.`mytab'.sep, top
	if "`dfdisplay'" == "dfdisplay" {
		.`mytab'.row .b .b "`e(vcetype)'" .b .b .b .b .b .b
		.`mytab'.titles "`e(depvar)'"               /// 1
					" Coefficient"            		/// 2
                    "std. err."        				/// 3
					"t"								/// 4
					"df"							/// 5
                    "scale"                        	/// 6
                    "P>|t|"                    		/// 7
                    "[`level'% conf. interval]" ""  //  8 9
	}
	else {
		.`mytab'.row .b .b "`e(vcetype)'" .b .b .b .b 
		.`mytab'.titles "`e(depvar)'"               /// 1
                     " Coefficient"            		/// 2
					"std. err."        				/// 3
                    "t"                        		/// 4
                    "P>|t|"                    		/// 5
                    "[`level'% conf. interval]" ""  //  6 7
	}
	.`mytab'.sep	
		
		
	local first 	// empty local for _ms_display
		
	forvalues i = 1/`k' {
		local name : word `i' of `namelist'	// extract variable
				
				
		_ms_display, el(`i') novbar `first' `diopts'
				
		if r(output) {
			local first
		}
		else {
			if r(first){
				local first first
			}
			continue
		}
			
		local note `"`r(note)'"'
		local err 0
			
		if "`note'" == "(empty)" {
			local err 1
			local err_lab "   (empty)  "
		}
		else if "`note'" == "(omitted)" {
			local err 1
			local err_lab "   (omitted)"
		}
			
			
		if `err' {
			if "`dfdisplay'" == "dfdisplay" {
				.`mytab'.row "" "." "`err_lab'" .b .b .b .b .b .b 
			}
			else {
				.`mytab'.row "" "." "`err_lab'" .b .b .b .b 
			}
					
		}
		else{
			// Computations and display
			scalar `scale' = `scmatrix'[1,`i']
			scalar `df' = `dfmatrix'[1,`i']
                
			scalar `t' = _b[`name']/_se[`name']
			scalar `p' = 2*ttail(`df', abs(`scale'*`t'))
			scalar `crit' = invttail(`df',(100-`level')/200)
			scalar `ll' = _b[`name']-`crit'*_se[`name']/`scale'
			scalar `ul' = _b[`name']+`crit'*_se[`name']/`scale'
				
			matrix `outmat'[1,`i'] = (_b[`name'] \ _se[`name'] \ `t' \ ///
										`p' \ `ll' \ `ul' \ `df' \ `crit' \ `scale') 
				
			if "`dfdisplay'" == "dfdisplay" {
				.`mytab'.row    ""          			///
								_b[`name']         		///
								_se[`name']        		///
								`t'						///
								`df'					///
								`scale'             	///
								`p'              		///
								`ll' `ul'
			}
			else {
				.`mytab'.row    ""         				///
								_b[`name']         		///
								_se[`name']        		///
								`t'             		///
								`p'              		///
								`ll' `ul'
			}
		}
				
	}
	.`mytab'.sep, bottom
	//.`mytab'.post_results, puttab(table) // investigate why this isn't working
		
	return matrix table = `outmat'
end




