#-----------------------------------------------------#
# Table3.R
#
# This program uses the R packages foreach, doParallel, extraDistr
#-----------------------------------------------------#
rep  <- 20000			# simulation replications
brep <- 999			# bootstrap replications
level <- .95			# confidence set level
Gs <- c(10,20,50,200)		# number of clusters
GTs <- c(4,3,2)			# number of treated clusters
nd <- 6				# number of designs
nm <- 8				# number of methods	
Ng <- 10			# average number of observations per cluster 
csizes <- 2			# csizes = 1 for equal cluster sizes
				#	 = 2 for unequal fixed cluster sizes
				#	 = 3 for random geometric cluster sizes
edist <- 1			# error: edist=1 for normal, edist=2 for Strongly Skewed, edist=3 for Binary
clusterfe <- 1			# clusterfe = 1 to include cluster fixed effects (estimated by within tranformation)
groupfe <- 0			# if groupfe > 0, include groupfe*G "large group" fixed effects, as dummy variables
extraregressors <- 2		# number of extra regressors included
T <- 2				# number of time periods
r <- 1				# coefficient of interest
seed <- 10232023		# random number seed 
Table = "Table3.txt"		# File for printed output
#-----------------------------------------------------#

### Moore-Penrose pseudoinverse for symmetric matrix A , raised to power p ###
Ginv <- function(A,p=1){
  tol <- 1e-15 * max(A) * nrow(A)
  ei <- eigen(A,symmetric=TRUE)
  d <- ei$values
  H <- ei$vectors
  n <- (d < tol)
  di <- 1/((d+n)^p)
  di[n] <- 0
  B <- H %*% (di*t(H))
  return(B)
}

### Strongly Skewed Distribution Random Number Generator ###
rStronglySkewed <- function(n){
  mu <- seq(-1,1,by=.25)*3/sqrt(12.75)
  sig <- sqrt((2^(1:9)))*9/sqrt(1022*12.75)
  i <- sample(1:9,n,replace=TRUE)
  e <- mu[i] + sig[i]*rnorm(n)
  return(e)
}

library(foreach)
library(doParallel)
cl <- makeCluster(nd)
registerDoParallel(cl)
start = Sys.time()

sink(Table)
cat("Replications:")
print(rep)
cat("Bootstrap Replications:")
print(brep)
cat("Nominal Coverage Level:")
print(level)
cat("Average Observations per Cluster:")
print(Ng)
cat("Number of treated clusters:")
print(GTs)
cat("Cluster Size Pattern:")
print(csizes)
cat("Error Distribution:")
print(edist)
if (clusterfe==1) cat("Cluster fixed effects included, estimated by within transformation\n")
if (groupfe>0) {
  cat("Large-group fixed effects included as dummy variables, proportional to G times:")
  print(groupfe)
}
cat("Number of extra regressors included:")
print(extraregressors)
cat("Number of time periods:")
print(T)
cat("Random Number Seed:")
print(seed)
cat("___________________")
sink()

for (gi in 1:length(Gs)){
G <- Gs[gi]
groups <- seq(G)

store = foreach(d=1:nd,.combine='rbind') %dopar% {
library(extraDistr)

sigtheta <- ifelse(d<=3,1,10)
GT <- ifelse(d<=3,GTs[d],GTs[d-3])
alpha <- ifelse(d<=3,.1,3)

storeA <- matrix(0,rep,nm)
set.seed(seed)  	
for (i in 1:rep) {

# Cluster sizes
if (csizes==1) 		ng <- rep(Ng,G)
else if (csizes==2) 	ng <- c(1+GT*(Ng-1),rep(1,GT-1),rep(Ng,G-GT))
else if (csizes==3) 	ng <- 1 + rgeom(G,1/Ng)
N <- sum(ng)			# number of individuals
group  <- rep(groups,T*ng)
n <- length(group)

# Create sample
time  <- rep((1:T),N)
treat <- (group <= GT)*(time > T/2)
X <- cbind(treat,time)
if (extraregressors > 0) {
  Z <- matrix(rnorm(n*extraregressors)+rep(treat,extraregressors),n,extraregressors)
  X <- cbind(X,Z)
}
if (groupfe > 0) {
  Dn <- groupfe*G
  D <- matrix(0,n,Dn)
  M <- 1/groupfe	# number of clusters per large group
  for (di in 1:Dn) {
    ind <- (group <= M*di)*(group > M*(di-1))
    D[,di] <- as.numeric(ind)
  }
  X <- cbind(X,D)
}
k <- ncol(X)
dof <- (G/(G-1)*(n-1)/(n-k))

h <- c(rep(1,ng[1]*T/2),rep(-1,ng[1]*T/2))
for (hi in 2:G) h <- c(h,c(rep(1,ng[hi]*T/2),rep(-1,ng[hi]*T/2)))

if (edist <= 2) {
  if (edist==1) {
    u <- rnorm(G)
    v <- rnorm(G)
    e <- rnorm(n)
    w <- rnorm(n)
  } else if (edist==2) {
    u <- rStronglySkewed(G)
    v <- rStronglySkewed(G)
    e <- rStronglySkewed(n)
    w <- rStronglySkewed(n)
  }
  theta0 <- 0
  theta <- theta0 + w*sigtheta
  Y0 <- e + u[group] + v[group]*h
  Y <- matrix(Y0 + treat*theta)
} else if (edist==3) {
  theta0 <- pnorm(alpha/sqrt(3)) - .5
  u <- rnorm(G)
  v <- rnorm(G)
  e <- rnorm(n)
  Y0 <- e + u[group] + v[group]*h
  Y <- matrix(as.numeric(Y0 > (1-treat)*alpha))
}

# Within Transformation
if (clusterfe == 1) {
  YX <- cbind(Y,X)        
  YX <- do.call(rbind, by(YX,group,function(x){sweep(x,2,colMeans(x),FUN = "-")})) 
  Y <- YX[,1] 
  X <- as.matrix(YX[,-1])
  dof <- (G/(G-1)*(n-1)/(n-k-G))
}

# Estimation
XX <- crossprod(X)
Q  <- solve(XX,tol=0)
XY <- crossprod(X,Y)
betahat <- Q%*%XY
ehat <- Y - X%*%betahat
thetahat <- betahat[r]
athetahat <- abs(thetahat - theta0)
Qr <- Q[,r]
rQ <- Q[r,]
Qrr <- Q[r,r]
betatilde <- betahat - Qr*((thetahat-theta0)/Qrr)

# SubCluster Calculations
U <- V <- Xe <- matrix(0,G,k)
A2 <- matrix(0,n,1)
GG <- matrix(0,G,G)
B <- matrix(0,n,G)
if (brep > 0) {
  XYs <- A <- F <- matrix(0,G,k)
  XXs <- Qs <- array(0,c(G,k,k))
}
betaJ <- S <- matrix(NA,G,1)
for (g in 1:G) {
  gi <- (group == g)
  ni <- sum(gi)
  Xg <- matrix(X[gi,],nrow=ni,ncol=k)
  Yg <- matrix(Y[gi],nrow=ni)
  XYg <- crossprod(Xg,Yg)
  XXg  <- crossprod(Xg)
  Qg <- Ginv(XX-XXg)
  betaJ[g] <- Qg[r,]%*%(XY-XYg) - thetahat 
  U[g,] <- Ug <- Qg%*%XXg%*%Qr
  V[g,] <- Vg <- XXg%*%(Qr + Ug)
  S[g]  <- t(Qr + Ug)%*%Vg
  Mg <- diag(ni) - Xg%*%Q%*%t(Xg)
  A2[gi] <- ag <- Ginv(Mg,.5)%*%Xg%*%Qr
  GG[g,g] <- sum(ag^2)
  B[,g] <- X%*%Q%*%crossprod(Xg,ag)
  Xe[g,] <- XYg - XXg%*%betahat
  if (brep > 0) {
    XYs[g,] <- XYg 
    XXs[g,,] <- XXg  
    Qs[g,,] <- Qg 
    A[g,]  <- XXg%*%Qr
    F[g,] <- XYg - XXg%*%betatilde
  }
}

# CRVE1
c1  <- qt((1+level)/2,G-1)
se1 <- sqrt(dof*rQ%*%crossprod(Xe)%*%Qr)
se1 <- se1[,]
storeA[i,1] <- (athetahat <= se1*c1)

# CRVE2 and Bell-McCaffrey critical value
se2  <- sqrt(sum(rowsum(A2*ehat,group,reorder=FALSE)^2))
storeA[i,2] <- (athetahat <= se2*c1)

# Bell-McCaffrey
GG <- GG - crossprod(B)
KBM <- (sum(diag(GG))^2)/sum(diag(GG%*%GG))
cBM <- qt((1+level)/2,KBM)
storeA[i,7] <- (athetahat <= se2*cBM)

# Jackknife
seJ <- sqrt(sum(betaJ^2))
storeA[i,4] <- (athetahat <= seJ*c1)

# Adjusted Jackknife
W <- sweep(U,1,S,FUN="*")
UU <- crossprod(U)
UV <- crossprod(U,V)
UW <- crossprod(U,W)
VV <- crossprod(V)
UUXX <- UU%*%XX
trL <- sum(S) + sum(diag(UUXX)) - 2*sum(diag(UV))
# The following uses the matrix fact:   sum(diag(A%*%B)) = sum(t(A)*B))
trLL <- sum(S^2) + sum(t(UUXX)*UUXX) + 2*sum(t(UV)*UV) + 2*sum(XX*UW) - 4*sum(W*V) - 4*sum(t(UUXX)*UV) + 2*sum(UU*VV)
KJack <- (trL^2)/trLL
aJack <- sqrt(trL/Qrr)
cJack <- qt((1+level)/2,KJack)/aJack
storeA[i,8] <- (athetahat <= seJ*cJack)

if (brep > 0) {

# Non-Parametric Pairs Bootstrap
storeb <- matrix(NA,brep,1)
for (bi in 1:brep) {
  groupb <- sample(1:G,G,replace=TRUE)
  XXb <- colSums(XXs[groupb,,],dims=1)
  if (rcond(XXb) > 1e-8) {
    XYb <- colSums(XYs[groupb,],dims=1)
    betahatb <- solve(XXb,XYb)
    storeb[bi] <- betahatb[r]
  }
}
storeb <- na.omit(storeb)
if (nrow(storeb) > 1)  storeA[i,3]  <- (athetahat <= sd(storeb)*c1)

# Wild Bootstrap
Xsi  <- matrix(rsign(G*brep),G,brep)
c1g <- F%*%Qr
thetahatb <- crossprod(c1g,Xsi)
FXsi <- crossprod(F,Xsi)
Z <- sweep(Xsi,1,c1g,"*") - A %*% Q %*% FXsi
se1b <- sqrt(colSums(Z^2)*dof)
p11 <- sum(se1b*athetahat < abs(thetahatb)*se1)/(brep+1)
storeA[i,5] <- (p11 > (1-level))

D  <- Qs[,,r]
h1  <- rowSums(F*D,dims=1)
V1b <- sweep(D%*%FXsi,2,thetahatb,"-") - sweep(Xsi,1,h1,"*")
seJb <- sqrt(colSums(V1b^2))
p12 <- sum(seJb*athetahat < abs(thetahatb)*seJ)/(brep+1)
storeA[i,6] <- (p12 > (1-level))

}

}	# end i

coverage <- colMeans(storeA)
coverage

}	# end d


sink(Table,append=TRUE)
cat("\nNumber of Groups G:")
print(G)
print.table(round(store,2))
sink()

}	# end g

sink(Table,append=TRUE)
print(Sys.time() - start)
sink()

stopCluster(cl)